--[[
Create visual elements.
Check items in DB against game.
Correct names, remove dead ones.


DToX, 2013-2014
Licensed under the same terms as Lua itself.
--]]





-- Variables and constants
--
local checkStartTime;
local checkEndTime;
local IDs; -- Used for loop, contains all IDs
local checkIndex; -- Index for loop
local namesFromGame; -- Pairs "<ID>" = "<Name>"
local outdatedNames; -- Pairs "<ID>" = { <NameFromDB>, <NameFromGame> }
local removedItems; -- IDs of the items, that are removed from game
local cntTotal;
local cntFound;
local cntSameNames;
local cntDifferentNames;
local LOOP_MAX_ITEMS = 100; -- Number of items to check in one loop
local REPORT_EVERY = 1000; -- Show status every N checked items (% operator)


-- Forward declarations
--
local qs;
local btn_checkDB;
local btn_updateNames;
local btn_removeItems;
local txt_result;





-- Functions
--
-- Checks item. Returns item's information
local function checkItem( decID ) -- Decimal ID. Example: 1879049233
   qs:SetShortcut( nil );
   local ok, sc = pcall( addShortcut, qs, decID );
   if ( ok ) then
      local info = sc:GetItem():GetItemInfo();
      return true, info;
   else
      return false;
   end
end

-- Sets initial values, enumerates IDs from DB, gets total number of items in DB
local function checkDB()
   txt_result:SetText( "Checking..." );
   IDs = {};
   namesFromGame = {};
   outdatedNames = {};
   removedItems = {};
   checkIndex = 1;
   cntFound = 0;
   cntSameNames = 0;
   cntDifferentNames = 0;

   -- Enumerate IDs. Get total number of items
   local items = data.Items;
   for ID in pairs( items ) do
      local i = #IDs + 1;
      IDs[i] = ID;
   end
   cntTotal = #IDs;
   if ( cntTotal == 0 ) then
      txt_result:SetText( "Nothing to check. DB is empty." );
      return;
   end

   -- Call qs:Update
   qs:SetWantsUpdates( true );
end

-- Counts found and not found items
-- Gets names of items from game, IDs of not found items
-- Checks for difference in names
local function checkItems()
   local i = 1;
   local items = data.Items;
   local ID = IDs[checkIndex];
   while ( ID and i <= LOOP_MAX_ITEMS ) do
      -- Check ID
      local ok, info = checkItem( ID );
      local strID = tostring( ID );
      if ( ok ) then
         -- Increase counter, save name
         cntFound = cntFound + 1;
         local name = info:GetName();
         namesFromGame[ID] = name;

         -- Compare names, increase counters. If different, save ID and names
         local nameFromDB = items[ID].Name;
         if ( nameFromDB ~= name ) then
            outdatedNames[strID] = {
               ["FromDB  "] = nameFromDB;
               ["FromGame"] = name;
            };
            cntDifferentNames = cntDifferentNames + 1;
         else
            cntSameNames = cntSameNames + 1;
         end
      else
         -- If not found, save ID
         local j = #removedItems + 1;
         removedItems[j] = strID;
      end

      -- Show progress and free memory
      if ( checkIndex % REPORT_EVERY == 0 ) then
         txt_result:SetText( "Checking...\nIDs checked: " .. checkIndex .. " (names found: " .. cntFound .. ")." );

         collectgarbage();
      end

      -- Increase loop's counter
      i = i + 1;
      checkIndex = checkIndex + 1;
      ID = IDs[checkIndex];

      -- No more items. Show status
      if ( not ID ) then
         -- Elapsed time
         checkEndTime = Turbine.Engine.GetLocalTime();
         local minutes = ( checkEndTime - checkStartTime ) / 60;
         minutes = string.format( "%.3f", minutes );
         txt_result:SetText( "Done. Elapsed minutes: " .. minutes .. "." );

         -- Status: items
         local cntNotFound = #removedItems;
         txt_result:AppendText( "\n[Items] Found + Not found = Sum / Total:\n"
         .. cntFound .. " + " .. cntNotFound .. " = " .. cntFound + cntNotFound .. " / " .. cntTotal .. "." );

         -- Status: names
         txt_result:AppendText( "\n[Names] Same + Different = Sum / Total (Total without Not found):\n"
         .. cntSameNames .. " + " .. cntDifferentNames .. " = " .. cntSameNames + cntDifferentNames .. " / " .. cntTotal .. " (" .. cntTotal - cntNotFound .. ")");

         -- Update results
         data.NotFound = removedItems;
         data.OutdatedNames = outdatedNames;

         -- Enable buttons or show message
         if ( cntDifferentNames > 0 ) then
            btn_updateNames:SetEnabled( true );
         else
            txt_result:AppendText( "\n\nNothing to update. Button stays disabled." );
         end
         if ( cntNotFound > 0 ) then
            btn_removeItems:SetEnabled( true );
         else
            txt_result:AppendText( "\n\nNothing to remove. Button stays disabled." );
         end

         return;
      end
   end

   qs:SetWantsUpdates( true );
end

-- Updates names in DB with names from game
local function updateNames()
   local items = data.Items;
   btn_checkDB:SetEnabled( false );
   txt_result:SetText( "Updating names..." );
   for ID in pairs( items ) do
      items[ID].Name = namesFromGame[ID];
   end
   txt_result:AppendText( " Done." );
   btn_checkDB:SetEnabled( true );
   btn_updateNames:SetEnabled( false );
end

-- Remove dead items from DB
local function removeDeadItems()
   local items = data.Items;
   btn_checkDB:SetEnabled( false );
   txt_result:SetText( "Removing items..." );
   for _, ID in pairs( removedItems ) do
      local numID = tonumber( ID );
      items[numID] = nil;
   end
   txt_result:AppendText( " Done." );
   btn_checkDB:SetEnabled( true );
   btn_removeItems:SetEnabled( false );
end



-- Visuals
--
-- Fills carrier with visual elements or just toggles visibility
-- Returns some elements for manipulation
local isCreated = false;
local correctWindowSize; -- Forward declaration for function
function showItems( mainCarrier )
   -- Main carrier for elements (offseted from window's borders)
   local mainCarrier = mainCarrier;

   if ( isCreated ) then
      mainCarrier:SetVisible( true );
      correctWindowSize();
      return;
   end

   local MAIN_CARRIER_WIDTH = 450;
   local MAIN_CARRIER_HEIGHT = 300;
   local MAIN_CARRIER_OFFSET = 25;
   local FONT = fonts.Verdana14;

   -- Corrects main window's size
   correctWindowSize = function()
      local NEW_MAIN_WINDOW_WIDTH = MAIN_CARRIER_WIDTH + MAIN_CARRIER_OFFSET * 2;
      local b = getBorders( mainCarrier );
      local newMainWindowHeight = MAIN_CARRIER_HEIGHT + MAIN_CARRIER_OFFSET + b.Top;
      win_main:SetSize( NEW_MAIN_WINDOW_WIDTH, newMainWindowHeight );
   end
   correctWindowSize();

   mainCarrier:SetSize( MAIN_CARRIER_WIDTH, MAIN_CARRIER_HEIGHT );

   -- Button: Check DB
   btn_checkDB = Turbine.UI.Lotro.Button();
   btn_checkDB:SetParent( mainCarrier );
   btn_checkDB:SetFont( FONT );
   btn_checkDB:SetText( "Check database" );
   btn_checkDB:SetSize( 130, 20 );
   btn_checkDB:SetPosition( 0, 0 );
   function btn_checkDB:MouseHover()
      Hint.show( self, "Scan game, collect data and show statistics (number of items, differences in names).");
   end
   function btn_checkDB:MouseClick()
      checkStartTime = Turbine.Engine.GetLocalTime();
      checkDB();
   end

   -- Button: Update names
   btn_updateNames = Turbine.UI.Lotro.Button();
   btn_updateNames:SetParent( mainCarrier );
   btn_updateNames:SetEnabled( false );
   btn_updateNames:SetFont( FONT );
   btn_updateNames:SetText( "Update names" );
   btn_updateNames:SetSize( 130, 20 );
   local b = getBorders( btn_checkDB )
   btn_updateNames:SetPosition( b.Left, b.Bottom + 10 );
   function btn_updateNames:MouseHover()
      Hint.show( self, "Replace names in DB with names from game.\nActual names can be found in *.plugindata file." );
   end
   function btn_updateNames:MouseClick()
      updateNames();
   end

   -- Button: Remove dead items
   btn_removeItems = Turbine.UI.Lotro.Button();
   btn_removeItems:SetParent( mainCarrier );
   btn_removeItems:SetEnabled( false );
   btn_removeItems:SetFont( FONT );
   btn_removeItems:SetText( "Remove items" );
   btn_removeItems:SetSize( 130, 20 );
   local b = getBorders( btn_updateNames )
   btn_removeItems:SetPosition( b.Right + 10 , b.Top );
   function btn_removeItems:MouseHover()
      Hint.show( self, "Remove dead items from DB.\nActual IDs can be found in *.plugindata file." );
   end
   function btn_removeItems:MouseClick()
      removeDeadItems();
   end

   -- Result: label
   local lbl_result = Turbine.UI.Label();
   lbl_result:SetParent( mainCarrier );
   lbl_result:SetFont( FONT );
   lbl_result:SetText( "Sanity check:" );
   lbl_result:SetSize( 200, 20 );
   local b = getBorders( btn_updateNames )
   lbl_result:SetPosition( b.Left, b.Bottom + 20 );

   -- Result: text
   txt_result = Turbine.UI.Lotro.TextBox();
   txt_result:SetParent( mainCarrier );
   txt_result:SetReadOnly( true );
   txt_result:SetFont( FONT );
   local b = getBorders( lbl_result );
   txt_result:SetSize( MAIN_CARRIER_WIDTH, MAIN_CARRIER_HEIGHT - b.Bottom );
   txt_result:SetPosition( b.Left, b.Bottom );

   -- Quickslot for checking items
   qs = Turbine.UI.Lotro.Quickslot();
   function qs:Update()
      self:SetWantsUpdates( false );
      checkItems();
   end

   -- Set flag, that elements are created, and return some elements for manipulation
   isCreated = true;
   return txt_result, btn_updateNames, btn_removeItems;
end