--[[
Create visual elements.
Show set's name, shortcuts to items, bonuses.
Save changes to set. Try to find same bonuses in other raw (unedited) sets and insert translation.
Correct saved set's information, when re-cheking current items.


DToX, 2013-2014
Licensed under the same terms as Lua itself.
--]]





-- Variables
--
local foundSets; -- Sets that are found from items data
local curBonusesPage = 1; -- Current bonuses' page
local curBonuses; -- Current set's bonuses
local bonusesMap; -- Pairs <Bonus> = <Translation>
local ignoreBonuses = ";Agility;All Skill Inductions;Armour;Audacity;Block Rating;Critical Defence;Critical Hit Avoidance;Critical Rating;Evade Chance;Evade Rating;Fate;Finesse Rating;in-Combat Morale Regen;in-Combat Power Regen;Incoming Healing Rating;Maximum Morale;Maximum Power;Might;non-Combat Morale Regen;non-Combat Power Regen;Parry Rating;Physical Mastery Rating;Physical Mitigation;Ranged Defence;Ranged Offence Rating;Resistance Rating;Stealth Level;Tactical Critical Multiplier;Tactical Mastery Rating;Tactical Mitigation;Vitality;Will;";
local listsMap; -- Pairs <Position in list> = <Set's index>
local clicked; -- String <Position in list><List's letter>. Example: 12R or 6E.
local editedBonusKeys; -- Contents of the popup list
local bugFix = false; -- To ignore some function when fixing sets
-- Quickslots: variables
local QUICKSLOT_WIDTH = 36;
local QUICKSLOT_HEIGHT = 36;
local QUICKSLOTS_NUMBER = 7;
local itemsQuickslots = {}; -- Holds quickslot objects
-- Bonuses: variables
local BONUS_WIDTH = 200;
local BONUS_HEIGHT = 100;
local BONUSES_PER_LINE = 3;
local BONUSES_LINES = 2;
local BONUSES_NUMBER = BONUSES_PER_LINE * BONUSES_LINES;
local BONUS_OUTER_OFFSET = 10; -- Between bonus' carriers
local BONUS_INNER_OFFSET = 5; -- Between bonus' elements
local BONUSES_WIDTH = BONUSES_PER_LINE * BONUS_WIDTH + ( BONUSES_PER_LINE - 1 ) * BONUS_OUTER_OFFSET;
-- Raw and edited sets: variables
local LIST_ITEM_HEIGHT = 13;
local LIST_ITEMS_NUMBER = 12;
local LIST_HEIGHT = LIST_ITEM_HEIGHT * LIST_ITEMS_NUMBER;
local LIST_WIDTH = 150;
local LIST_SCROLL_WIDTH = 12;
local LIST_ITEM_WIDTH = LIST_WIDTH - LIST_SCROLL_WIDTH;
-- Popup titles list: variables
local titlesList;
local LIST_SCROLL_WIDTH = 11;
local BORDER_WIDTH = 1;
local LIST_OFFSET = 5;
local TITLES_NUMBER = 10;
local TITLE_WIDTH = 180;
local TITLE_HEIGHT = 17;
local TITLES_WIDTH = TITLE_WIDTH + LIST_SCROLL_WIDTH;
local TITLES_CARRIER_WIDTH = TITLES_WIDTH + LIST_OFFSET * 2;
local TITLES_CARRIER_BACKGROUND_WIDTH = TITLES_CARRIER_WIDTH - BORDER_WIDTH * 2;
local MIN_TITLE_LENGTH_FOR_HINT = 28;


-- Forward declarations
--
local txt_setName;
local lst_bonusesControls;
local lst_bonuses;
local lbl_bonusesTotal;
local lst_rawSets;
local lst_editedSets;
local btn_incEquipped;
local btn_decEquipped;
local btn_updateSet;





-- Functions
--
local function getEditedBonusKeys()
   local sets = data.Sets;
   editedBonusKeys = {};

   for _, v in pairs( sets ) do
      if ( v.Set and v.Set.Bonus ) then
         local setBonuses = v.Set.Bonus;
         for _, iBonuses in pairs( setBonuses ) do
            for bonusKey in pairs( iBonuses ) do
               local j = #editedBonusKeys + 1;
               editedBonusKeys[j] = bonusKey;
            end
         end
      end
   end

   -- Remove duplicates and sort results
   editedBonusKeys = Table.dupes( editedBonusKeys, "Remove", true );
end

-- Shows object's possible values and sets selected
local function showTitlesList( obj, titles, mode )
   -- Exit if exists or list is empty
   local len = Table.len( titles );
   if ( titlesList or not len or ( len and len < 1 ) ) then
      return;
   end

   local objText = obj:GetText();
   local selectedItemIndex;
   local LCRL = { "а","б","в","г","д","е","ё","ж","з","и","й","к","л","м","н","о","п","р","с","т","у","ф","х","ц","ч","ш","щ","ъ","ы","ь","э","ю","я" };
   local UCRL = { "А","Б","В","Г","Д","Е","Ё","Ж","З","И","Й","К","Л","М","Н","О","П","Р","С","Т","У","Ф","Х","Ц","Ч","Ш","Щ","Ъ","Ы","Ь","Э","Ю","Я" };

   -- Forward declarations
   local carrierBackground;
   local veil;
   local lst_titles;
   local vScroll;
   local txt_listFilter;

   -- Closes titles window and hides veil
   local function closeTitles()
      if ( titlesList ) then
         titlesList:SetVisible( false );
         titlesList = nil;
      end
      if ( veil ) then
         veil:SetParent( nil );
         veil = nil;
      end
   end

   -- Show stat's hint if name is too long
   local function showTitleHint( obj )
      local length = obj:GetTextLength();
      if ( length > MIN_TITLE_LENGTH_FOR_HINT ) then
         local text = obj:GetText();
         Hint.show( obj, text );
      end
   end

   -- Adds items to the list, correct list's size and position
   local function fillTitles( titles )
      lst_titles:ClearItems();

      for i, name in pairs( titles ) do
         local name = tostring( name );
         local listItem = Turbine.UI.Label();
         listItem:SetMultiline( false );
         listItem:SetFont( fonts.Verdana14 );
         listItem:SetText( name );
         listItem:SetSize( TITLE_WIDTH, TITLE_HEIGHT );

         -- Highlight items and show hint for long names
         function listItem:MouseHover()
            listItem:SetBackColor( colors.Grey03 );
            showTitleHint( self );
         end
         -- Remove highlighting, but ignore highlighting of the selected item
         function listItem:MouseLeave()
            if ( name ~= objText ) then
               listItem:SetBackColor( nil );
            end
            Hint.remove();
         end
         -- Close titles window and change object's value
         function listItem:MouseClick()
            obj:SetText( name );
            closeTitles();
         end

         lst_titles:AddItem( listItem );

         -- Highlight selected item
         if ( name == objText ) then
            listItem:MouseHover();
            selectedItemIndex = i;
         end

         -- Correct list's and its scroll height
         if ( i <= TITLES_NUMBER ) then
            local h = i * TITLE_HEIGHT;
            lst_titles:SetHeight( h );
            vScroll:SetHeight( h );
         end

      end

      -- Scroll to the selected item
      lst_titles:EnsureVisible( selectedItemIndex );

      -- Correct titles window's height
      local h = lst_titles:GetHeight();
      h = h + LIST_OFFSET * 2;
      titlesList:SetHeight( h );
      carrierBackground:SetHeight( h - BORDER_WIDTH * 2 );

      -- Position titles window
      local left, top = getPosition( titlesList, obj, "Border", "MiddleRight", "Window" );
      left = left + TITLES_CARRIER_WIDTH / 4;
      local SCREEN_WIDTH = Turbine.UI.Display:GetWidth();
      if ( left + TITLES_CARRIER_WIDTH > SCREEN_WIDTH ) then
         left = SCREEN_WIDTH - TITLES_CARRIER_WIDTH - 5;
      end
      titlesList:SetPosition( left, top );

   end

   -- Converts russian letters to upper ones, otherwise uses standard function
   local function toUpper( str )
      local isRussian = false;
      local str = tostring( str );
      for i, letter in pairs ( LCRL ) do
         if ( str:find( letter ) ) then
            str = str:gsub( letter, UCRL[i] );
            isRussian = true;
         end
      end

      if ( not isRussian ) then
         str = string.upper( str );
      end

      return str;
   end


   -- Veil: hovers over main window. Used to close titles window
   veil = Turbine.UI.Control();
   veil:SetParent( win_main );
   veil:SetVisible( true );
   local w, h = win_main:GetSize();
   veil:SetSize( w, h );
   veil:Focus(); -- To remove focus from object
   -- Close titles window
   function veil:MouseClick()
      closeTitles();
   end
   win_main.PositionChanged = veil.MouseClick;
   -- Get focus to the filter field
   function veil:FocusGained()
      if ( titlesList ) then
         titlesList:Activate();
         txt_listFilter:Focus();
      end
   end

   -- Carrier: border's color
   titlesList = Turbine.UI.Window();
   titlesList:SetVisible( true );
   titlesList:SetBackColor( colors.Grey07 );
   titlesList:SetWidth( TITLES_CARRIER_WIDTH );
   -- Carrier: background
   carrierBackground = Turbine.UI.Control();
   carrierBackground:SetParent( titlesList );
   carrierBackground:SetBackColor( colors.Black );
   carrierBackground:SetWidth( TITLES_CARRIER_BACKGROUND_WIDTH );
   carrierBackground:SetPosition( BORDER_WIDTH, BORDER_WIDTH );

   -- Titles list
   lst_titles = Turbine.UI.ListBox();
   lst_titles:SetParent( titlesList );
   lst_titles:SetWidth( TITLES_WIDTH );
   lst_titles:SetPosition( LIST_OFFSET, LIST_OFFSET );
   -- List's scroll
   vScroll = Turbine.UI.Lotro.ScrollBar();
   vScroll:SetParent( lst_titles );
   vScroll:SetOrientation( Turbine.UI.Orientation.Vertical );
   vScroll:SetWidth( LIST_SCROLL_WIDTH );
   vScroll:SetPosition( TITLE_WIDTH, 0 );
   lst_titles:SetVerticalScrollBar( vScroll );

   -- Filter titles
   txt_listFilter = Turbine.UI.Lotro.TextBox();
   txt_listFilter:SetParent( titlesList );
   txt_listFilter:SetFont( fonts.Verdana14 );
   txt_listFilter:SetSize( 50, 20 );
   txt_listFilter:SetReadOnly( false );
   local left, top = getPosition( txt_listFilter, lst_titles, "Inside", "TopRight" );
   txt_listFilter:SetPosition( left, top );
   function txt_listFilter:TextChanged()
      local function cleanString( str )
         if ( str:find( "%W" ) ) then
            str = str:gsub( "%W", "_" );
         end
         return str;
      end

      local newTitles = {};

      -- Prepare inputed text
      local text = self:GetText();
      text = cleanString( text );
      text = toUpper( text );

      -- Search through all titles
      for i, title in pairs( titles ) do
         local upperCaseTitle = cleanString( title );
         upperCaseTitle = toUpper( upperCaseTitle );
         -- Compare title with inputed text and add, if matches
         if ( upperCaseTitle:find( text ) ) then
            local j = #newTitles + 1;
            newTitles[j] = title;
         end
      end

      if ( #newTitles < 1 ) then
         newTitles = titles;
      end

      fillTitles( newTitles );
   end

   fillTitles( titles );
end

-- Enables or disables pre-defined objects
local function toggleObjects( b )
   btn_incEquipped:SetEnabled( b );
   btn_decEquipped:SetEnabled( b );
   btn_updateSet:SetEnabled( b );
end

-- Sets equipped number, key and description of a bonus
-- at specified position in a list
local function setBonusValue( posNum, bonusNum, bonusDescr, bonusKey )
   local bCarrier = lst_bonuses:GetItem( posNum );
   local bFields = bCarrier:GetControls();
   if ( bonusNum ) then
      bFields:Get( 1 ):SetText( bonusNum ); -- 2
   end
   if ( bonusDescr ) then
      bFields:Get( 2 ):SetText( bonusDescr ); -- +15s
   end
   if ( bonusKey ) then
      bFields:Get( 3 ):SetText( bonusKey ); -- Duration
   end
end

-- Returns equipped number, key and description and carrier of a bonus
-- at specified position in a list
local function getBonusValue( posNum )
   local bCarrier = lst_bonuses:GetItem( posNum );
   local bFields = bCarrier:GetControls();
   local bNumber = bFields:Get( 1 ):GetText(); -- 2
   bNumber = tonumber( bNumber );
   local bDescr = bFields:Get( 2 ):GetText(); -- +15s
   local bKey = bFields:Get( 3 ):GetText(); -- Duration

   return bNumber, bDescr, bKey, bCarrier;
end

-- Clears set's fields and disables some buttons
local function clearFields()
   -- Clear set's name
   txt_setName:SetText( "" );
   -- Clear set's shortcuts to items
   local n = #itemsQuickslots;
   for i = 1, n do
      itemsQuickslots[i]:SetShortcut( nil );
   end
   -- Clear set's bonuses fields, clear current bonuses and select 1st page
   for i = 1, BONUSES_NUMBER do
      setBonusValue( i, "", "", "" )
   end
   curBonuses = {};
   lst_bonusesControls:GetItem( 1 ):GetControls():Get( 1 ):MouseClick();
   -- Disable buttons
   toggleObjects( false );
end

-- Fills fields with values or saves values from fields.
-- Updates number of bonuses
local function updateCurBonuses( mode )
   if ( not curBonuses ) then
      return;
   end

   local bNumStart = ( curBonusesPage - 1 ) * BONUSES_NUMBER + 1;

   for i = 1, BONUSES_NUMBER do

      -- Show bonus's values
      if ( mode == "Show" ) then
         -- Set carrier's text
         local _, _, _, bCarrier = getBonusValue( i );
         bCarrier:SetText( bNumStart );
         -- Clear carrier's subfields highlighting
         local bFields = bCarrier:GetControls();
         local color = nil;
         bFields:Get( 1 ):SetBackColor( color );
         bFields:Get( 2 ):SetBackColor( color );
         bFields:Get( 3 ):SetBackColor( color );

         local curBonus = curBonuses[bNumStart];
         if ( curBonus) then
            local bNumber, bKey, bDescr = curBonus.n, curBonus.k, curBonus.v;
            setBonusValue( i, bNumber, bDescr, bKey );

            -- Highlight subfields, if key matches ignored bonuses
            local key = bKey;
            if ( key:find( "-" ) ) then
               key = key:gsub( "-", "%%-" );
            end
            key = ";" .. key .. ";";
            if ( ignoreBonuses:match( key ) ) then
               color = colors.Grey03;
               bFields:Get( 1 ):SetBackColor( color );
               bFields:Get( 2 ):SetBackColor( color );
               bFields:Get( 3 ):SetBackColor( color );
            end
         else
            setBonusValue( i, "", "", "" );
         end

      -- Save values in array
      elseif ( mode == "Save" ) then
         local bNumber, bDescr, bKey = getBonusValue( i );
         curBonuses[bNumStart] = {
            ["n"] = bNumber or "";
            ["v"] = bDescr or "";
            ["k"] = bKey or "";
         };
      end

      bNumStart = bNumStart + 1;
   end

   -- Count bonuses
   local cnt = 0;
   for i, bonus in pairs( curBonuses ) do
      local n, k, v = bonus.n, bonus.k, bonus.v;
      if ( n ~= "" or k ~= "" or v ~= "" ) then
         cnt = cnt + 1;
      end
   end
   lbl_bonusesTotal:SetText( "Number of bonuses: " .. cnt .. "." );

   -- Change color of bonuses' number. Yellow = bonuses on another page
   local isAnotherPage = false;
   local page = 1;
   if ( curBonusesPage == 1 ) then
      page = 2;
   end
   local bNumStart = (page - 1) * BONUSES_NUMBER + 1;
   for i = 1, BONUSES_NUMBER do
      local b = curBonuses[bNumStart];
      if ( b and ( b.n ~= "" or b.v ~= "" or b.k ~= "" ) ) then
         isAnotherPage = true;
         break;
      end
      bNumStart = bNumStart + 1;
   end
   lbl_bonusesTotal:SetForeColor( colors.White );
   if ( cnt > 6 or isAnotherPage ) then
      lbl_bonusesTotal:SetForeColor( colors.Yellow );
   end
end

-- Creates current set's bonuses map. Updates all sets, that have same bonuses
local function setsBonuses( mode, bSilent )
   -- No bonuses, exit until
   if ( bonusesMap == 0 and mode ~= "Clear" ) then
      return;
   end

   if ( mode == "Clear" ) then
      bonusesMap = "";

   elseif ( mode == "Map" ) then
      local lastBonus = 0;
      local isOld = false;
      if ( bonusesMap == "" ) then
         isOld = true;
         bonusesMap = {};
      end

      for i, v in pairs( curBonuses ) do
         -- Ignore some keys
         local key = v.k;
         if ( key:find( "-" ) ) then
            key = key:gsub( "-", "%%-" );
         end
         key = ";" .. key .. ";";
         if ( v.k == "" or not v.k or ignoreBonuses:match( key ) ) then
            bonusesMap[i] = "";
         else
            local bonus = v.v .. "__" .. v.k;
            -- Fill map's Old and create New keys. Save last bonus' index
            if ( isOld ) then
               bonusesMap[i] = {
                  ["Old"] = bonus;
                  ["New"] = "";
               };
               lastBonus = i;
            -- Fill map's New keys
            elseif ( bonusesMap[i] and bonusesMap[i] ~= "" ) then
               bonusesMap[i]["New"] = bonus;
            end
         end
      end

      -- No bonuses
      if ( isOld and lastBonus == 0 ) then
         bonusesMap = 0;
      end

   -- Update all matches in raw sets' bonuses
   elseif ( mode == "Items" ) then
      -- Change structure from [<N>] = { [Old] = ""; [New] = "" } to [Old] = New
      local cntUniqueBonuses = 0;
      local tmp = {};
      for i, v in pairs( bonusesMap ) do
         if ( bonusesMap[i] ~= "" ) then
            local old = bonusesMap[i].Old;
            local new = bonusesMap[i].New;
            if ( not tmp[old] ) then
               cntUniqueBonuses = cntUniqueBonuses + 1;
            end
            tmp[old] = new;
         end
      end
      bonusesMap = tmp;

      -- Update sets' items
      local items = data.Items;
      local sets = data.Sets;
      local cntRawSetsChanged = 0;

      -- Loop through all raw sets
      local cntRawSets = 0;
      for _, setsData in pairs( sets ) do
      if ( setsData.isRaw ) then
         local ID = setsData.ID;
         local set = items[ID].Set;
         local setBonuses = set.Bonus;
         local isBonusFound = false;
         local newBonus = {};

         -- Get new bonus array
         -- Loop through all set's bonuses
         for eqNum, eqNumBonuses in pairs( setBonuses ) do
            newBonus[eqNum] = {};
         -- Loop through all bonuses for equipped number
         for k, v in pairs( eqNumBonuses ) do
            local rawBonus = v .. "__" .. k;
            local editedbonus = bonusesMap[rawBonus];
-- Bug fix. 2013.06.11. For correction of saved data. First version saved v as k and k as v
            local rawBonusOld = k .. "__" .. v;
            local editedbonusOld = bonusesMap[rawBonusOld];
            if ( not editedbonus and editedbonusOld ) then
               editedbonus = editedbonusOld;
            end
-- Bug fix. End
            if ( editedbonus ) then
               local delimStart, delimEnd = editedbonus:find( "__" );
               local editedBonusValue = editedbonus:sub( 1, delimStart - 1 );
               local editedBonusKey = editedbonus:sub( delimEnd + 1 );
               newBonus[eqNum][editedBonusKey] = editedBonusValue;
               isBonusFound = true;
            else
               newBonus[eqNum][k] = v;
            end
         end -- For. Bonuses for equipped number
         end -- For. All set's bonuses

         -- Replace old bonus array with new one in all set's items
         if ( isBonusFound ) then
            for i, iID in pairs( set.IDs ) do
               if ( items[iID] ) then
                  items[iID].Set.Bonus = newBonus;
               end
            end
            cntRawSetsChanged = cntRawSetsChanged + 1;
         end

         cntRawSets = cntRawSets + 1;
      end -- If isRaw == true
      end -- For. All sets

      if ( not bSilent ) then
         wl( ">> Unique bonuses mapped: " .. cntUniqueBonuses .. ". Raw sets changed: " .. cntRawSetsChanged .. "/" .. cntRawSets .. "." );
      end
   end

end

-- Fills set's properties (name, shortcuts, bonuses)
-- Shows error and IDs, if items are not in game
local function fillFields( item )
   -- Fill set's name
   local set = item.Set;
   local items = data.Items;
   txt_setName:SetText( set.Name );

   -- Fill quickslots with shortcuts to set's items.
   -- If there is no such item in game, show error and IDs.
   -- Ignore, if it's bug fix
   if ( not bugFix ) then

      local checkedItems = {
         ["Dead"] = {};
         ["NotInDB"] = {};
      };
      local n = #set.IDs;
      for i, ID in ipairs( set.IDs ) do
         if ( i > QUICKSLOTS_NUMBER ) then
            wl( ">> Showing only first " .. QUICKSLOTS_NUMBER .. " shortcuts to items (total: " .. n .. ")." );
            break;
         end

         local qs = itemsQuickslots[i];
         local ok = pcall( addShortcut, qs, ID ); -- !! Performance drop when changing sets
         if ( ok ) then
            if ( not items[ID] ) then
               local j = #checkedItems.NotInDB + 1;
               checkedItems.NotInDB[j] = ID;
            end
         else
            local j = #checkedItems.Dead + 1;
            checkedItems.Dead[j] = ID;
         end
      end
      if ( #checkedItems.Dead > 0 ) then
         wl( "Not found in game (" .. #checkedItems.Dead .. "/" .. n .. "): " .. table.concat( checkedItems.Dead, ", " ) .. ".", "Error" );
      end
      if ( #checkedItems.NotInDB > 0 ) then
         wl( "Not found in DB (" .. #checkedItems.NotInDB .. "/" .. n .. "): " .. table.concat( checkedItems.NotInDB, ", " ) .. ".", "Error" );
      end

   end

   -- Sort equipped numbers
   local sortedEqNum = {};
   for i in pairs( set.Bonus ) do
      local j = #sortedEqNum + 1;
      sortedEqNum[j] = i;
   end
   table.sort( sortedEqNum );
   -- Fill current bonuses array
   local bNum = 0;
   curBonuses = {};
   for _, n in pairs( sortedEqNum ) do
      for k, v in pairs( set.Bonus[n] ) do
         bNum = bNum + 1;
         curBonuses[bNum] = {
            ["n"] = n;
            ["v"] = v;
            ["k"] = k;
         };
      end
   end
   -- Fill bonuses fields
   updateCurBonuses( "Show" );
   -- Create map for bonuses
   setsBonuses( "Clear" );
   setsBonuses( "Map" );
end

-- Decreases or increases number of equipped items
-- in bonuses on current page by specified value
local function changeEqNum( way, val )
   local n = lst_bonuses:GetItemCount();
   for i = 1, n do
      local bNumber = getBonusValue( i );

      if ( bNumber ) then
         if ( way == "-" ) then
            bNumber = bNumber - val;
         elseif ( way == "+" ) then
            bNumber = bNumber + val;
         end

         if ( bNumber < 2 ) then
            bNumber = 2;
         elseif ( bNumber > 6 ) then
            bNumber = 6;
         end
         setBonusValue( i, bNumber );
      end
   end
end

-- When user clicks on item in lists
local function listItemClick( listName )
   clearFields();

   local list;
   local listMap;
   if ( listName == "R" ) then
      list = lst_rawSets;
      listMap = listsMap.Raw;
   else
      list = lst_editedSets;
      listMap = listsMap.Edited;
   end

   -- Remove highlighting from previous item
   if ( clicked ) then
      -- Correct list
      prevListName = clicked:match( "%a" );
      local list = list;
      if ( prevListName ~= listName ) then
         if ( prevListName == "R" ) then
            list = lst_rawSets;
         else
            list = lst_editedSets;
         end
      end
      -- Change color of the previous item
      local i = clicked:match( "%d+" );
      i = tonumber( i );
      local n = list:GetItemCount();
      if ( i <= n ) then
         local oldItem = list:GetItem( i );
         oldItem:SetBackColor( nil );
      end
   end
   -- Highlight selected item
   local listItem = list:GetSelectedItem();
   listItem:SetBackColor( colors.Grey03 );

   -- Save current position, fill fields, enable some buttons
   local items = data.Items;
   local sets = data.Sets;
   local i = list:GetSelectedIndex();
   clicked = i .. listName;
   local setIndex = listMap[i];
   local ID = sets[setIndex].ID;
   local item = items[ID];
   fillFields( item );
   toggleObjects( true );
end

-- Returns array with set's bonuses
local function getBonus()
   local setBonus = {};
   for i, bonus in pairs( curBonuses ) do
      local bNumber, bDescr, bKey = bonus.n, bonus.v, bonus.k;
      
      -- Fill array
      if ( bNumber ~= "" ) then
         if ( not setBonus[bNumber] ) then
            setBonus[bNumber] = {};
         end
         setBonus[bNumber][bKey] = bDescr;
      elseif ( bDescr ~= "" or bKey ~= "" ) then
         return false, i;
      end

   end

   return setBonus;
end

-- Saves changes and moves set from raw to edited list
local function updateSet()
   updateCurBonuses( "Save" );
   setsBonuses( "Map" );
   setsBonuses( "Items" );

   -- Check bonus. Return, if error
   local ok, blockNum = getBonus();
   if ( not ok ) then
      wl( "Block #" .. blockNum .. ". Please, enter bonus number or clear other fields too.", "Error" );
      return;
   end

   local listIndex = clicked:match( "%d+" );
   listIndex = tonumber( listIndex );
   local listName = clicked:match( "%a" );
   local list;
   local listMap;
   if ( listName == "R" ) then
      list = lst_rawSets;
      listMap = listsMap.Raw;
   else
      list = lst_editedSets;
      listMap = listsMap.Edited;
   end

   local items = data.Items;
   local sets = data.Sets;
   local setIndex = listMap[listIndex];
   local itemID = sets[setIndex].ID;
   local isRaw = sets[setIndex].isRaw;
   local IDs = items[itemID].Set.IDs;
   local setName = txt_setName:GetText();

   -- Replace set's name and bonuses in all items
   for _, ID in pairs( IDs ) do
      if ( items[ID] ) then
         local set = items[ID].Set;
         set.Name = setName;
         set.Bonus = getBonus();
      end
   end
   -- And save it in sets too
   sets[setIndex]["Set"] = items[itemID].Set;

   local listItem = list:GetItem( listIndex );
   listItem:SetBackColor( nil );
   listItem:SetText( setName );
   -- Move from raw list to edited. Correct listsMap
   if ( listName == "R" ) then
      local newList = lst_editedSets;
      local newListName = "E";
      local newMap = listsMap.Edited;

      -- Change MouseClick event
      function listItem:MouseClick()
         listItemClick( newListName );
      end

      -- Move to edited list, update listMap, change set's isRaw value
      list:RemoveItemAt( listIndex );
      newList:AddItem( listItem );
      local n = newList:GetItemCount();
      newMap[n] = listMap[listIndex];
      table.remove( listMap, listIndex );
      sets[setIndex].isRaw = false;

      -- Update index
      listIndex = n;
   end

   -- Select edited item and show bonuses
   lst_editedSets:SetSelectedIndex( listIndex );
   lst_editedSets:GetSelectedItem():MouseClick();

   -- Update available for selection bonuses
   getEditedBonusKeys();
end

-- Counts sets and sets' items. Fills array with sets
local function findSets()
   foundSets = {};
   local items = data.Items;
   local cntSetsItemsFound = 0; -- Number of items, that are parts of sets
   local tmpSetsItems = {}; -- Used to count sets, then destroyed

   -- Fill array with sets' items from full DB. Count items
   for ID, v in pairs( items ) do
      if ( v.Set ) then
         tmpSetsItems[ID] = Table.copy( v );
         cntSetsItemsFound = cntSetsItemsFound + 1;
      end
   end

   -- Fill sets array. Count items, that are in Set.IDs, but not in DB. Show IDs
   local cntNoItemInDB = 0;
   local noItemIDs = {}; -- String with IDs
   for ID, v in pairs( tmpSetsItems ) do
      for _, IDi in pairs( v.Set.IDs ) do
         -- If ID is not found, increase counter and add ID to message
         if ( not tmpSetsItems[IDi] ) then
            cntNoItemInDB = cntNoItemInDB + 1;
            local j = #noItemIDs + 1;
            noItemIDs[j] = IDi;
         end

         -- If ID in set equals ID of the item, add set
         -- Else Clear Set.IDs of set's other items
         if ( IDi == ID ) then
            local n = #foundSets + 1;
            foundSets[n] = {
               ["isRaw"] = true;
               ["ID"] = ID;
            };
         -- Clear Set.IDs of set's other items
         elseif ( tmpSetsItems[IDi] ) then
            tmpSetsItems[IDi].Set.IDs = {};
         end
      end
   end
   tmpSetsItems = nil;

   -- Show result
   local sets = data.Sets;
   local cntEditedSets = 0;
   for i, v in pairs( sets ) do
      if ( not v.isRaw ) then
         cntEditedSets = cntEditedSets + 1;
      end
   end
   local editedPercent = 0;
   if ( #foundSets > 0 or cntEditedSets > 0 ) then
      editedPercent = string.format( "%.2f", cntEditedSets / #foundSets * 100 );
   end
   wl( ">> Sets in DB: " .. #foundSets .. " (items: " .. cntSetsItemsFound .. "). Edited: " .. cntEditedSets .. " (" .. editedPercent .. "%)." );
   if ( cntNoItemInDB > 0 ) then
      wl( ">> Have ID in sets, but no items in DB (" .. cntNoItemInDB .. "): " .. table.concat( noItemIDs, ", " ) .. ".", "Error" );
   end

end

-- Merges saved sets with found ones.
-- If saved set has different ID, updates ID and IDs
local function mergeSets()
   local savedSets = data.Sets;

   if ( not savedSets ) then
      data.Sets = foundSets;
      return;
   end

   local foundSets = foundSets;
   local items = data.Items;
   local noItemInDB = {};

   -- Check saved sets and update corresponding raw ones
   for i, vSaved in pairs( savedSets ) do
      local found = false;
      local savedID = vSaved.ID;

      -- After initial compare data'll be containing raw sets too. Ignore them
      if ( not vSaved.isRaw ) then
         -- Check all found sets
         for j, vSets in pairs( foundSets ) do
            local ID = vSets.ID;

            -- Easy way: ID is found
            if ( ID == savedID ) then
               foundSets[j] = vSaved;
               foundSets[j].ID = ID;
               foundSets[j].Set.IDs = items[ID].Set.IDs;
               found = true;
               break;
            -- Hard way: ID isn't found, try all IDs
            else
               local IDs = items[ID].Set.IDs;
               for _, _ID in pairs( IDs ) do
                  if ( _ID == savedID ) then
                     foundSets[j] = vSaved;
                     foundSets[j].ID = ID;
                     foundSets[j].Set.IDs = items[ID].Set.IDs;
                     found = true;
                     break;
                  end
               end
               if ( found ) then
                  break;
               end
            end
         end

         if ( not found ) then
         -- !! Reverse look up from Saved set's ID -> Found set's ID -> Found set's IDs
         -- to Saved set's ID -> Saved set's IDs -> Found set's ID?
            local j = #noItemInDB + 1;
            noItemInDB[j] = savedID;
         end

      end

   end

   local n = #noItemInDB;
   if ( n > 0 ) then
      wl( "Can't find set's ID in DB as item (" .. n .. "): " .. table.concat( noItemInDB, ", " ) .. ".", "Error" );
   end

   data.Sets = foundSets;
end

-- Updates edited set's name and bonuses in items
local function updateSetsItems()
   local sets = data.Sets;
   local items = data.Items;

   for _, v in pairs( sets ) do
      if ( not v.isRaw ) then
         local setName = v.Set.Name;
         local setBonus = v.Set.Bonus;
         local IDs = v.Set.IDs;
         
         for _, ID in pairs( IDs ) do
            -- Ignore items, that are in sets' IDs, but not in DB
            if ( items[ID] ) then
               items[ID].Set.Name = setName;
               items[ID].Set.Bonus = setBonus;
            end
         end
      end
   end
end

-- Adds sets to the lists, creates listsMap
local function fillSetsLists()
   clearFields();
   lst_rawSets:ClearItems();
   lst_editedSets:ClearItems();

   listsMap = {
      ["Raw"] = {};
      ["Edited"] = {};
   };
   local items = data.Items;
   local sets = data.Sets;

   for i, v in pairs( sets ) do
      local list;
      local listName;
      local listMap;
      if ( v.isRaw ) then
         list = lst_rawSets;
         listName = "R";
         listMap = listsMap.Raw;
      else
         list = lst_editedSets;
         listName = "E";
         listMap = listsMap.Edited;
      end

      local listItem = Turbine.UI.Label();
      listItem:SetMultiline( false );
      listItem:SetFont( fonts.Verdana12 );
      listItem:SetText( items[v.ID].Set.Name );
      listItem:SetSize( LIST_ITEM_WIDTH, LIST_ITEM_HEIGHT );
      function listItem:MouseClick( args )
         -- Ignore right mouse button
         local button;
         if ( type( args ) == "table" ) then
            button = args.Button;
         else
            button = args;
         end
         if ( button == Turbine.UI.MouseButton.Right ) then
            return;
         end

         listItemClick( listName );
      end
      -- Highlighting
      function listItem:MouseEnter()
         self:SetBackColor( colors.Grey03 );
      end
      function listItem:MouseLeave()
         self:SetBackColor( nil );
         local index;
         if ( v.isRaw ) then
            index = lst_rawSets:IndexOfItem( self ) .. "R";
         else
            index = lst_editedSets:IndexOfItem( self ) .. "E";
         end
         if ( clicked and index == clicked ) then
            self:SetBackColor( colors.Grey03 );
         end
      end

      list:AddItem( listItem );

      local n = list:GetItemCount();
      listMap[n] = i;
   end
end

-- Compilation of functions
local function checkSets()
   wl( ">> Checking sets..." );
   findSets();
   mergeSets();
   updateSetsItems();
   fillSetsLists();
   clickedCheckSets = true;
   getEditedBonusKeys();
   collectgarbage();
end

-- Bug fix. 2013.06.11. Corects raw sets bonuses
local function fixRawSetsBonuses()
   bugFix = true;
   local cnt = 0;
   local cntRawSets = lst_rawSets:GetItemCount();
   if ( cntRawSets > 0 ) then
      local sets = data.Sets;

      for i, v in pairs( sets ) do
         if ( not v.isRaw ) then
            cnt = cnt + 1;
            lst_editedSets:SetSelectedIndex( cnt )
            lst_editedSets:GetItem( cnt ):MouseClick(); -- !! Game freezer
            setsBonuses( "Map" );
            setsBonuses( "Items", "Silent" );
         end
      end
   end
   lst_editedSets:SetSelectedIndex( 0 );
   clearFields();
   wl( ">> Edited sets re-mapped: " .. cnt .. "." );
   bugFix = false;
end
-- Bug fix. End





-- Visuals
--
-- Fills carrier with visual elements or just toggles visibility
-- Returns some elements for manipulation
local isCreated = false;
local correctWindowSize; -- Forward declaration for function
function showSets( mainCarrier )
   -- Main carrier for elements (offseted from window's borders)
   local mainCarrier = mainCarrier;

   if ( isCreated ) then
      mainCarrier:SetVisible( true );
      correctWindowSize();
      return;
   end

   local FONT = fonts.Verdana14;

   -- Quickslots: functions
   local function addQuickslot()
      local qs = Turbine.UI.Lotro.Quickslot();
      qs:SetSize( QUICKSLOT_WIDTH, QUICKSLOT_HEIGHT );
      return qs;
   end

   -- Bonuses: functions
   -- Returns one bonus object with corrected offsets
   local function addBonus()
      -- Offset next bonus
      local startLeft = 0;
      local startTop = 0;
      local n = lst_bonuses:GetItemCount() + 1;
      if ( n % BONUSES_PER_LINE ~= 1 ) then
         startLeft = 10;
      end
      if ( n > BONUSES_PER_LINE ) then
         startTop = 10;
      end

      -- Carrier: holds one bonus
      local bonusCarrier = Turbine.UI.Label();

      -- Text: Number of equipped items
      local txt_bonusNumber = Turbine.UI.Lotro.TextBox();
      txt_bonusNumber:SetParent( bonusCarrier );
      txt_bonusNumber:SetMultiline( false );
      txt_bonusNumber:SetTextAlignment( align[5] );
      txt_bonusNumber:SetFont( FONT );
      txt_bonusNumber:SetSize( 20, 20 );
      txt_bonusNumber:SetPosition( startLeft, startTop );
      -- Number. Check input. Accept only 1 digit from 2 to 6
      function txt_bonusNumber:TextChanged()
         local inputed = txt_bonusNumber:GetText();
         local digit = inputed:match( "[2-6]" );
         txt_bonusNumber:SetText( digit ); -- !! Почему-то сдвигает введенное чуток вниз
      end

      -- Text: DB's value. Example: "-15s", "Chance", etc.
      local txt_bonusDescr = Turbine.UI.Lotro.TextBox();
      txt_bonusDescr:SetParent( bonusCarrier );
      txt_bonusDescr:SetMultiline( false );
      txt_bonusDescr:SetTextAlignment( align[4] );
      txt_bonusDescr:SetFont( FONT );
      local b = getBorders( txt_bonusNumber );
      local objW = txt_bonusNumber:GetWidth();
      local w = BONUS_WIDTH - b.Right - BONUS_INNER_OFFSET * 2 - objW + startLeft;
      txt_bonusDescr:SetSize( w, 20 );
      txt_bonusDescr:SetPosition( b.Right + BONUS_INNER_OFFSET, b.Top );

      -- Text: DB's key. Example: ["Will"], ["Test of Will Cooldown"], etc.
      local txt_bonusKey = Turbine.UI.Lotro.TextBox();
      txt_bonusKey:SetParent( bonusCarrier );
      txt_bonusKey:SetMultiline( true );
      txt_bonusKey:SetFont( FONT );
      txt_bonusKey:SetBackColorBlendMode( 7 );
      local b = getBorders( txt_bonusNumber );
      local h = BONUS_HEIGHT - b.Bottom - BONUS_INNER_OFFSET + startTop;
      txt_bonusKey:SetSize( BONUS_WIDTH, h );
      txt_bonusKey:SetPosition( b.Left, b.Bottom + BONUS_INNER_OFFSET );

      -- Button: select key from the list of edited ones
      local lbl_selectBonusKey = Turbine.UI.Label();
      lbl_selectBonusKey:SetParent( bonusCarrier );
      lbl_selectBonusKey:SetFont( FONT );
      lbl_selectBonusKey:SetTextAlignment( align[5] );
      lbl_selectBonusKey:SetText( "S" );
      lbl_selectBonusKey:SetSize( 17, 20 );
      lbl_selectBonusKey:SetBackColor( colors.Grey03 );
      local left, top = getPosition( lbl_selectBonusKey, txt_bonusKey, "Inside", "BottomRight" );
      lbl_selectBonusKey:SetPosition( left, top );
      function lbl_selectBonusKey:MouseClick()
         showTitlesList( txt_bonusKey, editedBonusKeys );
      end

      -- Button: clear fields
      -- Text field. Provides border
      local txt_clearBonus = Turbine.UI.Lotro.TextBox();
      txt_clearBonus:SetParent( bonusCarrier );
      txt_clearBonus:SetSize( 20, 20 );
      local b = getBorders( txt_bonusDescr );
      txt_clearBonus:SetPosition( b.Right + BONUS_INNER_OFFSET, b.Top );
      
      local btn_clearBonus = Turbine.UI.Label();
      btn_clearBonus:SetParent( bonusCarrier );
      btn_clearBonus:SetBackColor( colors.Grey03 );
      btn_clearBonus:SetFontStyle( Turbine.UI.FontStyle.Outline );
      btn_clearBonus:SetOutlineColor( colors.Grey07 );
      btn_clearBonus:SetTextAlignment( align[5] );
      btn_clearBonus:SetFont( FONT );
      btn_clearBonus:SetText( "C" );
      btn_clearBonus:SetSize( 18, 18 );
      local b = getBorders( txt_clearBonus );
      btn_clearBonus:SetPosition( b.Left + 1, b.Top + 1 );
      function btn_clearBonus:MouseHover()
         Hint.show( self, "Clear all subfields (equipped number, bonus's key, bonus's description)." );
      end
      -- Clear subfields and back color
      function btn_clearBonus:MouseClick()
         txt_bonusNumber:SetText( "" );
         txt_bonusKey:SetText( "" );
         txt_bonusDescr:SetText( "" );
         updateCurBonuses( "Save" );
         updateCurBonuses( "Show" );
      end


      -- Set size to bonus' carrier, write its number and return it
      local b = getBorders( txt_bonusKey );
      bonusCarrier:SetSize( b.Right, b.Bottom );
      bonusCarrier:SetTextAlignment( align[8] );
      bonusCarrier:SetFont( fonts.Verdana23 );
      bonusCarrier:SetForeColor( colors.Grey07 );
      bonusCarrier:SetText( n );
      return bonusCarrier;
   end

   -- Bonuses' page controls: variables
   local BONUSES_CONTROL_WIDTH = 40;
   local BONUSES_CONTROL_HEIGHT = 20;
   local BONUSES_CONTROL_OFFSET_H = 10; -- Between controls
   -- Bonuses' controls:  functions
   -- Returns button, that switches set's equipped bonuses
   local function addBonusesControl( num )
      local left = 0;
      if ( num > 1 ) then
         left = BONUSES_CONTROL_OFFSET_H;
      end

      local carrier = Turbine.UI.Control();
      local w = BONUSES_CONTROL_WIDTH + left;
      carrier:SetWidth( w );
      local control = Turbine.UI.Lotro.Button();
      control:SetTextAlignment( align[5] );
      control:SetParent( carrier );
      control:SetFont( FONT );
      control:SetText( num );
      control:SetSize( BONUSES_CONTROL_WIDTH, BONUSES_CONTROL_HEIGHT );
      control:SetLeft( left );
      -- For animation
      -- When button is disabled and mouse enters it, animation restarts
      -- So, well... veil
      local veil = Turbine.UI.Control();
      veil:SetParent( carrier );
      veil:SetSize( control:GetSize() );
      veil:SetLeft( left );
      veil:SetVisible( false );

      function control:MouseClick()
         -- Save bonuses
         updateCurBonuses( "Save" );

         -- Enable current/Disable previous button, hide/show veil over it
         -- Update current page number vatiable
         local prevControlNum = tonumber( curBonusesPage );
         local controlCarrier = lst_bonusesControls:GetItem( prevControlNum );
         local prevControl = controlCarrier:GetControls():Get( 1 );
         local prevVeil = controlCarrier:GetControls():Get( 2 );
         prevVeil:SetVisible( false );
         prevControl:SetEnabled( true );
         veil:SetVisible( true );
         self:SetEnabled( false );
         curBonusesPage = self:GetText();
         curBonusesPage = tonumber( curBonusesPage );

         -- Show bonuses
         updateCurBonuses( "Show" );
      end

      return carrier;
   end


   -- Button: check sets
   local btn_checkSets = Turbine.UI.Lotro.Button();
   btn_checkSets:SetParent( mainCarrier );
   btn_checkSets:SetFont( FONT );
   btn_checkSets:SetText( "Check sets" );
   btn_checkSets:SetSize( 100, 20 );
   function btn_checkSets:MouseHover()
      Hint.show( self, "Check sets in DB against saved ones, update saved sets and items. Fill lists. Show statistics." );
   end
   function btn_checkSets:MouseClick()
      checkSets();
   end

-- Bug fix. 2013.06.11. Button: fix raw sets' bonuses
   local btn_fixRawSets = Turbine.UI.Lotro.Button();
   btn_fixRawSets:SetParent( mainCarrier );
   btn_fixRawSets:SetFont( FONT );
   btn_fixRawSets:SetText( "Fix raw bonuses" );
   btn_fixRawSets:SetSize( 130, 20 );
   function btn_fixRawSets:MouseHover()
      Hint.show( self, "First public version of the plugin incorrectly updated raw sets' bonuses. Press this button to correct such bonuses.\n\nWarning! Takes some time. Game will not respond until plugin is done." );
   end
   function btn_fixRawSets:MouseClick()
      if ( not clickedCheckSets ) then
         checkSets();
      end
      fixRawSetsBonuses();
   end
-- Bug fix. End

   -- Button: Clear fields
   local btn_clearFields = Turbine.UI.Lotro.Button();
   btn_clearFields:SetParent( mainCarrier );
   btn_clearFields:SetFont( FONT );
   btn_clearFields:SetText( "Clear fields" );
   btn_clearFields:SetSize( 90, 20 );
   function btn_clearFields:MouseHover()
      Hint.show( self, "Clear set's name, set's shortcuts to items, set's bonuses.\nNo changes to the data." );
   end
   function btn_clearFields:MouseClick()
      clearFields();
   end


   -- Label: control equipped number
   local lbl_equippedNum = Turbine.UI.Label();
   lbl_equippedNum:SetParent( mainCarrier );
   lbl_equippedNum:SetTextAlignment( align[4] );
   lbl_equippedNum:SetFont( FONT );
   lbl_equippedNum:SetText( "Equipped number:" );
   lbl_equippedNum:SetSize( 120, 20 );

   local str = " number of equipped items for all set's bonuses by one.";
   -- Button: increase equipped number
   btn_incEquipped = Turbine.UI.Lotro.Button();
   btn_incEquipped:SetParent( mainCarrier );
   btn_incEquipped:SetFont( FONT );
   btn_incEquipped:SetText( "+1" );
   btn_incEquipped:SetSize( 40, 20 );
   function btn_incEquipped:MouseHover()
      Hint.show( self, "Increase" .. str );
   end
   function btn_incEquipped:MouseClick()
      changeEqNum( "+", 1 );
   end

   -- Button: decrease equipped number
   btn_decEquipped = Turbine.UI.Lotro.Button();
   btn_decEquipped:SetParent( mainCarrier );
   btn_decEquipped:SetFont( FONT );
   btn_decEquipped:SetText( "-1" );
   btn_decEquipped:SetSize( 40, 20 );
   function btn_decEquipped:MouseHover()
      Hint.show( self, "Decrease" .. str );
   end
   function btn_decEquipped:MouseClick()
      changeEqNum( "-", 1 );
   end


   -- Button: Save set
   btn_updateSet = Turbine.UI.Lotro.Button();
   btn_updateSet:SetParent( mainCarrier );
   btn_updateSet:SetFont( FONT );
   btn_updateSet:SetText( "Update set" );
   btn_updateSet:SetSize( 100, 20 );
   function btn_updateSet:MouseHover()
      Hint.show( self, "Update all set's items. Move to edited list." );
   end
   function btn_updateSet:MouseClick()
      updateSet();
   end

   -- Set's name text field
   txt_setName = Turbine.UI.Lotro.TextBox();
   txt_setName:SetMultiline( true );
   txt_setName:SetTextAlignment( align[4] );
   txt_setName:SetFont( FONT );
   local w = QUICKSLOT_WIDTH * QUICKSLOTS_NUMBER;
   txt_setName:SetSize( w, 40 );
   -- Add border and label to the element
   local setNameCarrier = addBorder( "Set's name", txt_setName );
   setNameCarrier:SetParent( mainCarrier );
   -- Select text and gain focus
   function setNameCarrier:MouseEnter()
      local t = txt_setName:GetText();
      if ( t == "" ) then
         return;
      end
      win_main:Activate();
      txt_setName:Focus();
      txt_setName:SelectAll();
   end
   function setNameCarrier:MouseLeave()
      win_main:Focus();
   end

   -- List: Holds quickslots with shortcuts to set's items
   local lst_quickslots = Turbine.UI.ListBox();
   lst_quickslots:SetOrientation( Turbine.UI.Orientation.Horizontal );
   lst_quickslots:SetSize( QUICKSLOT_WIDTH * QUICKSLOTS_NUMBER, QUICKSLOT_HEIGHT );
   -- Add border and label to the element
   local quickslotsCarrier = addBorder( "Set's items", lst_quickslots );
   quickslotsCarrier:SetParent( mainCarrier );
   -- Add quickslots objects to the list and to the array for easier access
   for i = 1, QUICKSLOTS_NUMBER do
      local qs = addQuickslot();
      lst_quickslots:AddItem( qs );
      itemsQuickslots[i] = qs;
   end


   -- Bonuses
   -- Temp carrier for page controls and bonuses
   local tmpCarrier = Turbine.UI.Control();

   -- Bonuses' controls. Label
   local lbl_bonusesControls = Turbine.UI.Label();
   lbl_bonusesControls:SetParent( tmpCarrier );
   lbl_bonusesControls:SetTextAlignment( align[4] );
   lbl_bonusesControls:SetFont( FONT );
   lbl_bonusesControls:SetText( "Pages:" );
   lbl_bonusesControls:SetSize( 50, 20 );

   -- Bonuses' controls. List: holds bonuses' controls
   lst_bonusesControls = Turbine.UI.ListBox();
   lst_bonusesControls:SetParent( tmpCarrier );
   lst_bonusesControls:SetOrientation( Turbine.UI.Orientation.Horizontal );
   -- Bonuses' controls. Add controls
   for i = 1, 2 do
      local bc = addBonusesControl( i );
      lst_bonusesControls:AddItem( bc );
   end
   -- Bonuses' controls. Calculate and set list's size
   local n = lst_bonusesControls:GetItemCount();
   local w = BONUSES_CONTROL_WIDTH * n + BONUSES_CONTROL_OFFSET_H * (n - 1);
   lst_bonusesControls:SetSize( w, BONUSES_CONTROL_HEIGHT );
   local b = getBorders( lbl_bonusesControls );
   lst_bonusesControls:SetLeft( b.Right + 10 );

   -- Bonuses. Label: total number
   lbl_bonusesTotal = Turbine.UI.Label();
   lbl_bonusesTotal:SetParent( tmpCarrier );
   lbl_bonusesTotal:SetTextAlignment( align[4] );
   lbl_bonusesTotal:SetFont( FONT );
   lbl_bonusesTotal:SetText( "Number of bonuses: 0" );
   lbl_bonusesTotal:SetSize( 160, 20 );
   local b = getBorders( lst_bonusesControls );
   lbl_bonusesTotal:SetLeft( b.Right + 30 );
   function lbl_bonusesTotal:MouseHover()
      Hint.show( self, "Number of bonuses, that are currently filled (equipped number, key or value).\nUpdates number when page changes or bonus is cleared." );
   end

   -- List: holds bonuses
   lst_bonuses = Turbine.UI.ListBox();
   lst_bonuses:SetParent( tmpCarrier );
   lst_bonuses:SetOrientation( Turbine.UI.Orientation.Horizontal );
   lst_bonuses:SetMaxItemsPerLine( BONUSES_PER_LINE );
   local h = BONUS_HEIGHT * 2 + BONUS_OUTER_OFFSET;
   lst_bonuses:SetSize( BONUSES_WIDTH, h );
   local b = getBorders( lbl_bonusesControls );
   lst_bonuses:SetPosition( b.Left, b.Bottom + 10 );
   -- Add bonus fields
   for i = 1, BONUSES_NUMBER do
      lst_bonuses:AddItem( addBonus() );
   end

   -- Add border and label to the element and position it
   local b = getBorders( lst_bonuses );
   tmpCarrier:SetSize( b.Right, b.Bottom  );
   local bonusesCarrier = addBorder( "Set's bonuses", tmpCarrier );
   bonusesCarrier:SetParent( mainCarrier );


   -- List: holds raw sets
   lst_rawSets = Turbine.UI.ListBox();
   lst_rawSets:SetSize( LIST_WIDTH, LIST_HEIGHT );
   -- Raw sets: vertical scroll
   local vScroll = Turbine.UI.Lotro.ScrollBar();
   vScroll:SetParent( lst_rawSets );
   vScroll:SetOrientation( Turbine.UI.Orientation.Vertical );
   vScroll:SetSize( LIST_SCROLL_WIDTH, LIST_HEIGHT );
   vScroll:SetLeft( LIST_ITEM_WIDTH );
   lst_rawSets:SetVerticalScrollBar( vScroll );
   -- Raw sets: add border and label to the element
   local rawSetsCarrier = addBorder( "Raw sets", lst_rawSets );
   rawSetsCarrier:SetParent( mainCarrier );

   -- List: holds edited sets
   lst_editedSets = Turbine.UI.ListBox();
   lst_editedSets:SetSize( LIST_WIDTH, LIST_HEIGHT );
   -- Edited sets: vertical scroll
   local vScroll = Turbine.UI.Lotro.ScrollBar();
   vScroll:SetParent( lst_editedSets );
   vScroll:SetOrientation( Turbine.UI.Orientation.Vertical );
   vScroll:SetSize( LIST_SCROLL_WIDTH, LIST_HEIGHT );
   vScroll:SetLeft( LIST_ITEM_WIDTH );
   lst_editedSets:SetVerticalScrollBar( vScroll );
   -- Edited sets: add border and label to the element
   local editedSetsCarrier = addBorder( "Edited sets", lst_editedSets );
   editedSetsCarrier:SetParent( mainCarrier );


   -- Position all elements
   btn_checkSets:SetPosition( 0, 0 );
   -- Position raw and edited lists
   local b = getBorders( bonusesCarrier );
   local objW = editedSetsCarrier:GetWidth();
   local left = b.Right - objW;
   editedSetsCarrier:SetPosition( left, 0 );
   local b = getBorders( editedSetsCarrier );
   local objW = rawSetsCarrier:GetWidth();
   local left = b.Left - objW - 20;
   rawSetsCarrier:SetPosition( left, 0 );
   -- Position set's quickslots
   local b = getBorders( rawSetsCarrier );
   local objH = quickslotsCarrier:GetHeight();
   local top = b.Bottom - objH;
   quickslotsCarrier:SetPosition( 0, top );
   -- Position set's name
   local b = getBorders( quickslotsCarrier );
   local objH = setNameCarrier:GetHeight();
   local top = b.Top - objH - 10;
   setNameCarrier:SetPosition( b.Left, top );
-- Bug fix. 2013.06.11. Position fix raw sets button
   local b = getBorders( btn_checkSets );
   local top = b.Top;
   local b = getBorders( setNameCarrier );
   local objW = btn_fixRawSets:GetWidth();
   local left = b.Right - objW;
   btn_fixRawSets:SetPosition( left, top );
-- Bug fix. End
   -- Position set's bonuses
   local b = getBorders( quickslotsCarrier );
   bonusesCarrier:SetPosition( b.Left, b.Bottom + 10 );
   -- Position bottom line: clear fields, change equipped number, update set
   local b = getBorders( bonusesCarrier );
   btn_clearFields:SetPosition( b.Left, b.Bottom + 15 );
   local b = getBorders( btn_clearFields );
   lbl_equippedNum:SetPosition( b.Right + 20, b.Top );
   local b = getBorders( lbl_equippedNum );
   btn_incEquipped:SetPosition( b.Right + 10, b.Top );
   local b = getBorders( btn_incEquipped );
   btn_decEquipped:SetPosition( b.Right + 10, b.Top );
   local b = getBorders( bonusesCarrier );
   local objW = btn_updateSet:GetWidth();
   local left = b.Right - objW;
   btn_updateSet:SetPosition( left, b.Bottom + 15 );

   -- Set carrier's size
   local b = getBorders( btn_updateSet );
   local mainCarrierWidth = b.Right;
   local mainCarrierHeight = b.Bottom;
   mainCarrier:SetSize( mainCarrierWidth, mainCarrierHeight );

   -- Corrects main window's size
   correctWindowSize = function()
      local MAIN_CARRIER_OFFSET = 25;
      local newMainWindowWidth = mainCarrierWidth + MAIN_CARRIER_OFFSET * 2;
      local b = getBorders( mainCarrier );
      local newMainWindowHeight = mainCarrierHeight + MAIN_CARRIER_OFFSET + b.Top;
      win_main:SetSize( newMainWindowWidth, newMainWindowHeight );
   end
   correctWindowSize();

   clearFields();

   -- Set flag, that elements are created, and return some elements for manipulation
   isCreated = true;
   return btn_checkSets;
end